@@ -1,5 +1,5 @@
 /* cms.c - cryptographic message syntax main functions
- *      Copyright (C) 2001, 2003, 2004 g10 Code GmbH
+ *      Copyright (C) 2001, 2003, 2004, 2008 g10 Code GmbH
  *
  * This file is part of KSBA.
  *
@@ -65,16 +65,16 @@ static struct {
   { NULL }
 };
 
-static char oidstr_contentType[] = "1.2.840.113549.1.9.3";
+static const char oidstr_contentType[] = "1.2.840.113549.1.9.3";
 /*static char oid_contentType[9] = "\x2A\x86\x48\x86\xF7\x0D\x01\x09\x03";*/
 
-static char oidstr_messageDigest[] = "1.2.840.113549.1.9.4";
-static char oid_messageDigest[9] = "\x2A\x86\x48\x86\xF7\x0D\x01\x09\x04";
+static const char oidstr_messageDigest[] = "1.2.840.113549.1.9.4";
+static const char oid_messageDigest[9] ="\x2A\x86\x48\x86\xF7\x0D\x01\x09\x04";
 
-static char oidstr_signingTime[] = "1.2.840.113549.1.9.5";
-static char oid_signingTime[9] = "\x2A\x86\x48\x86\xF7\x0D\x01\x09\x05";
+static const char oidstr_signingTime[] = "1.2.840.113549.1.9.5";
+static const char oid_signingTime[9] = "\x2A\x86\x48\x86\xF7\x0D\x01\x09\x05";
 
-static char oidstr_smimeCapabilities[] = "1.2.840.113549.1.9.15";
+static const char oidstr_smimeCapabilities[] = "1.2.840.113549.1.9.15";
 
 
 
@@ -1329,7 +1329,7 @@ ksba_cms_set_content_type (ksba_cms_t cms, int what, ksba_content_type_t type)
  * Set the algorithm to be used for creating the hash. Note, that we
  * currently can't do a per-signer hash.
  * 
- * Return value: o on success or an error code
+ * Return value: 0 on success or an error code
  **/
 gpg_error_t
 ksba_cms_add_digest_algo (ksba_cms_t cms, const char *oid)
@@ -2258,8 +2258,14 @@ store_smime_capability_sequence (AsnNode node,
         }
       if (cap2 == cap)
         {
+          /* RFC3851 requires that a missing parameter must not be
+             encoded as NULL.  This is in contrast to all other usages
+             of the algorithm identifier where ist is allowed and in
+             some profiles (e.g. tmttv2) even explicitly suggested to
+             use NULL.  */
           err = _ksba_der_write_algorithm_identifier
-                 (tmpwrt, cap->oid, cap->parmlen?cap->parm:NULL, cap->parmlen);
+                 (tmpwrt, cap->oid, 
+                  cap->parmlen?cap->parm:(const void*)"", cap->parmlen);
           if (err)
             {
               ksba_writer_release (tmpwrt);
@@ -2398,6 +2404,14 @@ build_signed_data_attributes (ksba_cms_t cms)
       unsigned char *image;
       size_t imagelen;
 
+      for (i = 0; i < attridx; i++)
+        {
+          _ksba_asn_release_nodes (attrarray[i].root);
+          xfree (attrarray[i].image);
+        }
+      attridx = 0;
+      memset (attrarray, 0, sizeof (attrarray));
+
       if (!digestlist)
         {
 	  err = gpg_error (GPG_ERR_MISSING_VALUE); /* oops */
@@ -2580,6 +2594,7 @@ build_signed_data_attributes (ksba_cms_t cms)
 	  goto leave;
 	}
 
+      assert (attridx <= DIM (attrarray));
       for (i=0; i < attridx; i++)
         {
           if (i)
@@ -2620,7 +2635,7 @@ build_signed_data_attributes (ksba_cms_t cms)
   for (i = 0; i < attridx; i++)
     {
       _ksba_asn_release_nodes (attrarray[i].root);
-      free (attrarray[i].image);
+      xfree (attrarray[i].image);
     }
 
   return err;
