@@ -1,5 +1,5 @@
 /* keyinfo.c - Parse and build a keyInfo structure
- *      Copyright (C) 2001, 2002 g10 Code GmbH
+ *      Copyright (C) 2001, 2002, 2007, 2008 g10 Code GmbH
  *
  * This file is part of KSBA.
  *
@@ -27,7 +27,6 @@
 #include <string.h>
 #include <assert.h>
 
-#include "ksba.h"
 #include "util.h"
 #include "asn1-func.h"
 #include "keyinfo.h"
@@ -35,142 +34,172 @@
 #include "convert.h"
 #include "ber-help.h"
 
+
+/* Constants used for the public key algorithms.  */
+typedef enum 
+  {
+    PKALGO_RSA,
+    PKALGO_DSA,
+    PKALGO_ECC
+  }
+pkalgo_t;
+
+
 struct algo_table_s {
   const char *oidstring;
   const unsigned char *oid;  /* NULL indicattes end of table */
   int                  oidlen;
   int supported;
-  int is_ecc;
+  pkalgo_t pkalgo;
   const char *algo_string;
   const char *elem_string; /* parameter name or '-' */
   const char *ctrl_string; /* expected tag values (value > 127 are raw data)*/
-  int digest_algo;
+  const char *parmelem_string; /* parameter name or '-'. */
+  const char *parmctrl_string; /* expected tag values.  */
+  const char *digest_string; /* The digest algo if included in the OID. */
 };
 
-static struct algo_table_s pk_algo_table[] = {
+
+
+static const struct algo_table_s pk_algo_table[] = {
 
   { /* iso.member-body.us.rsadsi.pkcs.pkcs-1.1 */
     "1.2.840.113549.1.1.1", /* rsaEncryption (RSAES-PKCA1-v1.5) */ 
     "\x2a\x86\x48\x86\xf7\x0d\x01\x01\x01", 9, 
-    1, 0, "rsa", "-ne", "\x30\x02\x02" },
+    1, PKALGO_RSA, "rsa", "-ne", "\x30\x02\x02" },
 
   { /* iso.member-body.us.rsadsi.pkcs.pkcs-1.7 */
     "1.2.840.113549.1.1.7", /* RSAES-OAEP */ 
     "\x2a\x86\x48\x86\xf7\x0d\x01\x01\x07", 9, 
-    0, 0, "rsa", "-ne", "\x30\x02\x02"}, /* (patent problems) */
+    0, PKALGO_RSA, "rsa", "-ne", "\x30\x02\x02"}, /* (patent problems) */
 
   { /* */
     "2.5.8.1.1", /* rsa (ambiguous due to missing padding rules)*/
     "\x55\x08\x01\x01", 4, 
-    1, 0, "ambiguous-rsa", "-ne", "\x30\x02\x02" },
+    1, PKALGO_RSA, "ambiguous-rsa", "-ne", "\x30\x02\x02" },
 
   { /* iso.member-body.us.x9-57.x9cm.1 */
     "1.2.840.10040.4.1", /*  dsa */
     "\x2a\x86\x48\xce\x38\x04\x01", 7, 
-    1, 0, "dsa", "y", "\x02" }, 
-  /* FIXME: Need code to extract p,q,g from the parameters */
+    1, PKALGO_DSA, "dsa", "y", "\x02", "-pqg", "\x30\x02\x02\x02" }, 
  
   { /* iso.member-body.us.ansi-x9-62.2.1 */
     "1.2.840.10045.2.1", /*  ecPublicKey */
     "\x2a\x86\x48\xce\x3d\x02\x01", 7, 
-    1, 1,"ecc", "q", "\x80" }, 
+    1, PKALGO_ECC, "ecc", "q", "\x80" }, 
 
   {NULL}
 };
 
 
-static struct algo_table_s sig_algo_table[] = {
+static const struct algo_table_s sig_algo_table[] = {
   {  /* iso.member-body.us.rsadsi.pkcs.pkcs-1.5 */
     "1.2.840.113549.1.1.5", /* sha1WithRSAEncryption */ 
     "\x2A\x86\x48\x86\xF7\x0D\x01\x01\x05", 9, 
-    1, 0, "rsa", "s", "\x82", GCRY_MD_SHA1 },
+    1, PKALGO_RSA, "rsa", "s", "\x82", NULL, NULL, "sha1" },
   { /* iso.member-body.us.rsadsi.pkcs.pkcs-1.4 */
     "1.2.840.113549.1.1.4", /* md5WithRSAEncryption */ 
     "\x2A\x86\x48\x86\xF7\x0D\x01\x01\x04", 9, 
-    1, 0, "rsa", "s", "\x82", GCRY_MD_MD5 },
+    1, PKALGO_RSA, "rsa", "s", "\x82", NULL, NULL, "md5" },
   { /* iso.member-body.us.rsadsi.pkcs.pkcs-1.2 */
     "1.2.840.113549.1.1.2", /* md2WithRSAEncryption */ 
     "\x2A\x86\x48\x86\xF7\x0D\x01\x01\x02", 9, 
-    0, 0, "rsa", "s", "\x82", 0 },
+    0, PKALGO_RSA, "rsa", "s", "\x82", NULL, NULL, "md2" },
+  { /* iso.member-body.us.x9-57.x9cm.1 */
+    "1.2.840.10040.4.3", /* dsa */
+    "\x2a\x86\x48\xce\x38\x04\x01", 7, 
+    1, PKALGO_DSA, "dsa", "-rs", "\x30\x02\x02" }, 
   { /* iso.member-body.us.x9-57.x9cm.3 */
     "1.2.840.10040.4.3", /*  dsaWithSha1 */
     "\x2a\x86\x48\xce\x38\x04\x03", 7, 
-    1, 0, "dsa", "-rs", "\x30\x02\x02", GCRY_MD_SHA1 }, 
+    1, PKALGO_DSA, "dsa", "-rs", "\x30\x02\x02", NULL, NULL, "sha1" }, 
+  { /* Teletrust signature algorithm.  */
+    "1.3.36.8.5.1.2.2", /* dsaWithRIPEMD160 */
+    "\x06\x07\x2B\x24\x08\x05\x01\x02\x02", 9,
+    1, PKALGO_DSA, "dsa", "-rs", "\x30\x02\x02", NULL, NULL, "rmd160" }, 
+  { /* NIST Algorithm */
+    "2.16.840.1.101.3.4.3.1", /* dsaWithSha224 */
+    "\x06\x09\x60\x86\x48\x01\x65\x03\x04\x03\x01", 11,
+    1, PKALGO_DSA, "dsa", "-rs", "\x30\x02\x02", NULL, NULL, "sha224" }, 
+  { /* NIST Algorithm (the draft also used .1 but we better use .2) */
+    "2.16.840.1.101.3.4.3.2", /* dsaWithSha256 */
+    "\x06\x09\x60\x86\x48\x01\x65\x03\x04\x03\x01", 11,
+    1, PKALGO_DSA, "dsa", "-rs", "\x30\x02\x02", NULL, NULL, "sha256" }, 
 
   { /* iso.member-body.us.ansi-x9-62.signatures.ecdsa-with-sha1 */
     "1.2.840.10045.4.1", /*  ecdsa */
     "\x2a\x86\x48\xce\x3d\x04\x01", 7, 
-    1, 1, "ecdsa", "-rs", "\x30\x02\x02", GCRY_MD_SHA1 }, 
+    1, PKALGO_ECC, "ecdsa", "-rs", "\x30\x02\x02", NULL, NULL, "sha1" }, 
 
   { /* iso.member-body.us.ansi-x9-62.signatures.ecdsa-with-specified */
     "1.2.840.10045.4.3", 
     "\x2a\x86\x48\xce\x3d\x04\x03", 7, 
-    1, 1, "ecdsa", "-rs", "\x30\x02\x02", 0 }, 
+    1, PKALGO_ECC, "ecdsa", "-rs", "\x30\x02\x02", NULL, NULL, NULL }, 
   /* The digest algorithm is given by the parameter.  */ 
 
 
   { /* iso.member-body.us.ansi-x9-62.signatures.ecdsa-with-sha224 */
     "1.2.840.10045.4.3.1", 
     "\x2a\x86\x48\xce\x3d\x04\x03\x01", 8, 
-    1, 1, "ecdsa", "-rs", "\x30\x02\x02", GCRY_MD_SHA224 }, 
+    1, PKALGO_ECC, "ecdsa", "-rs", "\x30\x02\x02", NULL, NULL, "sha224" }, 
 
   { /* iso.member-body.us.ansi-x9-62.signatures.ecdsa-with-sha256 */
     "1.2.840.10045.4.3.2", 
     "\x2a\x86\x48\xce\x3d\x04\x03\x02", 8, 
-    1, 1, "ecdsa", "-rs", "\x30\x02\x02", GCRY_MD_SHA256 }, 
+    1, PKALGO_ECC, "ecdsa", "-rs", "\x30\x02\x02", NULL, NULL, "sha256" }, 
 
   { /* iso.member-body.us.ansi-x9-62.signatures.ecdsa-with-sha384 */
     "1.2.840.10045.4.3.3", 
     "\x2a\x86\x48\xce\x3d\x04\x03\x03", 8, 
-    1, 1, "ecdsa", "-rs", "\x30\x02\x02", GCRY_MD_SHA384 }, 
+    1, PKALGO_ECC, "ecdsa", "-rs", "\x30\x02\x02", NULL, NULL, "sha384" }, 
 
   { /* iso.member-body.us.ansi-x9-62.signatures.ecdsa-with-sha512 */
     "1.2.840.10045.4.3.4", 
     "\x2a\x86\x48\xce\x3d\x04\x03\x04", 8, 
-    1, 1, "ecdsa", "-rs", "\x30\x02\x02", GCRY_MD_SHA512 }, 
+    1, PKALGO_ECC, "ecdsa", "-rs", "\x30\x02\x02", NULL, NULL, "sha512" }, 
 
   { /* iso.member-body.us.rsadsi.pkcs.pkcs-1.1 */
     "1.2.840.113549.1.1.1", /* rsaEncryption used without hash algo*/ 
     "\x2a\x86\x48\x86\xf7\x0d\x01\x01\x01", 9, 
-    1, 0, "rsa", "s", "\x82" },
+    1, PKALGO_RSA, "rsa", "s", "\x82" },
   { /* from NIST's OIW - actually belongs in a pure hash table */
     "1.3.14.3.2.26",  /* sha1 */
     "\x2B\x0E\x03\x02\x1A", 5,
-    0, 0, "sha-1", "", "", GCRY_MD_SHA1 },
+    0, PKALGO_RSA, "sha-1", "", "", NULL, NULL, "sha1" },
 
   { /* As used by telesec cards */
     "1.3.36.3.3.1.2",  /* rsaSignatureWithripemd160 */
     "\x2b\x24\x03\x03\x01\x02", 6,
-    1, 0, "rsa", "s", "\x82", GCRY_MD_RMD160 },
+    1, PKALGO_RSA, "rsa", "s", "\x82", NULL, NULL, "rmd160" },
 
   { /* from NIST's OIW - used by TU Darmstadt */
     "1.3.14.3.2.29",  /* sha-1WithRSAEncryption */
     "\x2B\x0E\x03\x02\x1D", 5,
-    1, 0, "rsa", "s", "\x82", GCRY_MD_SHA1 },
+    1, PKALGO_RSA, "rsa", "s", "\x82", NULL, NULL, "sha1" },
 
   { /* from PKCS#1  */
     "1.2.840.113549.1.1.11", /* sha256WithRSAEncryption */
     "\x2a\x86\x48\x86\xf7\x0d\x01\x01\x0b", 9,
-    1, 0, "rsa", "s", "\x82", GCRY_MD_SHA256 },
+    1, PKALGO_RSA, "rsa", "s", "\x82", NULL, NULL, "sha256" },
 
   { /* from PKCS#1  */
     "1.2.840.113549.1.1.12", /* sha384WithRSAEncryption */
     "\x2a\x86\x48\x86\xf7\x0d\x01\x01\x0c", 9,
-    1, 0, "rsa", "s", "\x82", GCRY_MD_SHA384 },
+    1, PKALGO_RSA, "rsa", "s", "\x82", NULL, NULL, "sha384" },
 
   { /* from PKCS#1  */
     "1.2.840.113549.1.1.13", /* sha512WithRSAEncryption */
     "\x2a\x86\x48\x86\xf7\x0d\x01\x01\x0d", 9,
-    1, 0, "rsa", "s", "\x82", GCRY_MD_SHA512 },
+    1, PKALGO_RSA, "rsa", "s", "\x82", NULL, NULL, "sha512" },
 
   {NULL}
 };
 
-static struct algo_table_s enc_algo_table[] = {
+static const struct algo_table_s enc_algo_table[] = {
   { /* iso.member-body.us.rsadsi.pkcs.pkcs-1.1 */
     "1.2.840.113549.1.1.1", /* rsaEncryption (RSAES-PKCA1-v1.5) */ 
     "\x2A\x86\x48\x86\xF7\x0D\x01\x01\x01", 9, 
-    1, 0, "rsa", "a", "\x82" },
+    1, PKALGO_RSA, "rsa", "a", "\x82" },
   {NULL}
 };
 
@@ -220,10 +249,10 @@ struct stringbuf {
 };
 
 
-#define TLV_LENGTH() do {         \
-  if (!derlen)                    \
+#define TLV_LENGTH(prefix) do {         \
+  if (!prefix ## len)                    \
     return gpg_error (GPG_ERR_INV_KEYINFO);  \
-  c = *der++; derlen--;           \
+  c = *(prefix)++; prefix ## len--;           \
   if (c == 0x80)                  \
     return gpg_error (GPG_ERR_NOT_DER_ENCODED);  \
   if (c == 0xff)                  \
@@ -238,32 +267,17 @@ struct stringbuf {
       for (len=0; count; count--) \
         {                         \
           len <<= 8;              \
-          if (!derlen)            \
+          if (!prefix ## len)            \
             return gpg_error (GPG_ERR_BAD_BER);\
-          c = *der++; derlen--;   \
+          c = *(prefix)++; prefix ## len--;   \
           len |= c & 0xff;        \
         }                         \
     }                             \
-  if (len > derlen)               \
+  if (len > prefix ## len)               \
     return gpg_error (GPG_ERR_INV_KEYINFO);  \
 } while (0)
 
 
-#if 0
-static void
-dump_hex (const unsigned char *p, size_t n)
-{
-  if (!p)
-    fputs (" none", stderr);
-  else
-    {
-      for (; n; n--, p++)
-        fprintf (stderr, " %02X", *p);
-    }
-}
-#endif
-
-
 /* Given a string BUF of length BUFLEN with either the name of an ECC
    curve or its OID in dotted form return the DER encoding of the OID.
    The caller must free the result.  On error NULL is returned.  */
@@ -328,7 +342,7 @@ get_algorithm (int mode, const unsigned char *der, size_t derlen,
   c = *der++; derlen--;
   if ( c != 0x30 )
     return gpg_error (GPG_ERR_UNEXPECTED_TAG); /* not a SEQUENCE */
-  TLV_LENGTH(); 
+  TLV_LENGTH(der); 
   seqlen = len;
   startseq = der;
 
@@ -338,7 +352,7 @@ get_algorithm (int mode, const unsigned char *der, size_t derlen,
   c = *der++; derlen--; 
   if ( c != 0x06 )
     return gpg_error (GPG_ERR_UNEXPECTED_TAG); /* not an OBJECT IDENTIFIER */
-  TLV_LENGTH();
+  TLV_LENGTH(der);
 
   /* der does now point to an oid of length LEN */
   *r_pos = der - start;
@@ -376,7 +390,7 @@ get_algorithm (int mode, const unsigned char *der, size_t derlen,
           /*  This is an octet string parameter and we need it.  */
           if (r_parm_type)
             *r_parm_type = TYPE_OCTET_STRING;
-          TLV_LENGTH();
+          TLV_LENGTH(der);
           *r_parm_pos = der - start;
           *r_parm_len = len;
           seqlen -= der - startparm;
@@ -389,7 +403,7 @@ get_algorithm (int mode, const unsigned char *der, size_t derlen,
           /*  This is an object identifier.  */
           if (r_parm_type)
             *r_parm_type = TYPE_OBJECT_ID;
-          TLV_LENGTH();
+          TLV_LENGTH(der);
           *r_parm_pos = der - start;
           *r_parm_len = len;
           seqlen -= der - startparm;
@@ -402,7 +416,7 @@ get_algorithm (int mode, const unsigned char *der, size_t derlen,
           /*  This is a sequence. */
           if (r_parm_type)
             *r_parm_type = TYPE_SEQUENCE;
-          TLV_LENGTH();
+          TLV_LENGTH(der);
           *r_parm_pos = startparm - start;
           *r_parm_len = len + (der - startparm);
           seqlen -= der - startparm;
@@ -413,7 +427,7 @@ get_algorithm (int mode, const unsigned char *der, size_t derlen,
       else
         {
 /*            printf ("parameter: with tag %02x - ignored\n", c); */
-          TLV_LENGTH();
+          TLV_LENGTH(der);
           seqlen -= der - startparm;
           /* skip the value */
           der += len;
@@ -438,7 +452,7 @@ get_algorithm (int mode, const unsigned char *der, size_t derlen,
         ; /* OCTECT STRING */
       else
         return gpg_error (GPG_ERR_UNEXPECTED_TAG); /* not a BIT STRING */
-      TLV_LENGTH();
+      TLV_LENGTH(der);
     }
 
   *r_nread = der - start;
@@ -612,8 +626,6 @@ get_stringbuf (struct stringbuf *sb)
                     parameters   ANY DEFINED BY algorithm OPTIONAL }
                  publicKey  BIT STRING }
   
-  We only allow parameters == NULL.
-
   The function parses this structure and create a SEXP suitable to be
   used as a public key in Libgcrypt.  The S-Exp will be returned in a
   string which the caller must free.  
@@ -631,6 +643,8 @@ _ksba_keyinfo_to_sexp (const unsigned char *der, size_t derlen,
   char *parm_oid = NULL;
   int algoidx;
   int is_bitstr;
+  const unsigned char *parmder = NULL;
+  size_t parmderlen = 0;
   const unsigned char *ctrl;
   const char *elem;
   struct stringbuf sb;
@@ -643,7 +657,7 @@ _ksba_keyinfo_to_sexp (const unsigned char *der, size_t derlen,
   c = *der++; derlen--;
   if ( c != 0x30 )
     return gpg_error (GPG_ERR_UNEXPECTED_TAG); /* not a SEQUENCE */
-  TLV_LENGTH();
+  TLV_LENGTH(der);
   /* and now the inner part */
   err = get_algorithm (1, der, derlen, &nread, &off, &len, &is_bitstr,
                        &parm_off, &parm_len, &parm_type);
@@ -662,9 +676,13 @@ _ksba_keyinfo_to_sexp (const unsigned char *der, size_t derlen,
   if (!pk_algo_table[algoidx].supported)
     return gpg_error (GPG_ERR_UNSUPPORTED_ALGORITHM);
 
-
   if (parm_off && parm_len && parm_type == TYPE_OBJECT_ID)
     parm_oid = ksba_oid_to_str (der+parm_off, parm_len);
+  else if (parm_off && parm_len)
+    {
+      parmder = der + parm_off;
+      parmderlen = parm_len;
+    }
 
   der += nread;
   derlen -= nread;
@@ -694,7 +712,7 @@ _ksba_keyinfo_to_sexp (const unsigned char *der, size_t derlen,
   put_stringbuf_sexp (&sb, pk_algo_table[algoidx].algo_string);
 
   /* Insert the curve name for ECC. */
-  if (pk_algo_table[algoidx].is_ecc && parm_oid)
+  if (pk_algo_table[algoidx].pkalgo == PKALGO_ECC && parm_oid)
     {
       put_stringbuf (&sb, "(");
       put_stringbuf_sexp (&sb, "curve");
@@ -702,6 +720,56 @@ _ksba_keyinfo_to_sexp (const unsigned char *der, size_t derlen,
       put_stringbuf (&sb, ")");
     }
 
+  /* If parameters are given and we have a description for them, parse
+     them. */
+  if (parmder && parmderlen 
+      && pk_algo_table[algoidx].parmelem_string
+      && pk_algo_table[algoidx].parmctrl_string)
+    {
+      elem = pk_algo_table[algoidx].parmelem_string; 
+      ctrl = pk_algo_table[algoidx].parmctrl_string; 
+      for (; *elem; ctrl++, elem++)
+        {
+          int is_int;
+          
+          if ( (*ctrl & 0x80) && !elem[1] )
+            {
+              /* Hack to allow reading a raw value.  */
+              is_int = 1;
+              len = parmderlen;
+            }
+          else
+            {
+              if (!parmderlen)
+                {
+                  xfree (parm_oid);
+                  return gpg_error (GPG_ERR_INV_KEYINFO);
+                }
+              c = *parmder++; parmderlen--;
+              if ( c != *ctrl )
+                {
+                  xfree (parm_oid);
+                  return gpg_error (GPG_ERR_UNEXPECTED_TAG);
+                }
+              is_int = c == 0x02;
+              TLV_LENGTH (parmder);
+            }
+          if (is_int && *elem != '-')  /* Take this integer.  */
+            { 
+              char tmp[2];
+              
+              put_stringbuf (&sb, "(");
+              tmp[0] = *elem; tmp[1] = 0;
+              put_stringbuf_sexp (&sb, tmp);
+              put_stringbuf_mem_sexp (&sb, parmder, len);
+              parmder += len;
+              parmderlen -= len;
+              put_stringbuf (&sb, ")");
+            }
+        }
+    }
+
+
   /* FIXME: We don't release the stringbuf in case of error
      better let the macro jump to a label */
   elem = pk_algo_table[algoidx].elem_string; 
@@ -730,7 +798,7 @@ _ksba_keyinfo_to_sexp (const unsigned char *der, size_t derlen,
               return gpg_error (GPG_ERR_UNEXPECTED_TAG);
             }
           is_int = c == 0x02;
-          TLV_LENGTH ();
+          TLV_LENGTH (der);
         }
       if (is_int && *elem != '-')  /* Take this integer.  */
         { 
@@ -758,18 +826,14 @@ _ksba_keyinfo_to_sexp (const unsigned char *der, size_t derlen,
 
 /* Match the algorithm string given in BUF which is of length BUFLEN
    with the known algorithms from our table and returns the table
-   entries for the DER encoded OID.
-
-   FIXME: We restrict this for now to RSA and ECC because the code
-   using this function is not yet prepared to handle other
-   algorithms.  */
+   entries for the DER encoded OID.  */
 static const unsigned char *
 oid_from_buffer (const unsigned char *buf, int buflen, int *oidlen,
-                 int *r_is_ecc)
+                 pkalgo_t *r_pkalgo)
 {
   int i;
 
-  /* ignore a leading "oid." string */
+  /* Ignore an optional "oid." prefix. */
   if (buflen > 4 && buf[3] == '.' && digitp (buf+4)
       && ((buf[0] == 'o' && buf[1] == 'i' && buf[2] == 'd')
           ||(buf[0] == 'O' && buf[1] == 'I' && buf[2] == 'D')))
@@ -778,7 +842,7 @@ oid_from_buffer (const unsigned char *buf, int buflen, int *oidlen,
       buflen -= 4;
     }
 
-  /* and scan the table */
+  /* Scan the table. */
   for (i=0; pk_algo_table[i].oid; i++)
     {
       if (!pk_algo_table[i].supported)
@@ -792,12 +856,8 @@ oid_from_buffer (const unsigned char *buf, int buflen, int *oidlen,
     }
   if (!pk_algo_table[i].oid)
     return NULL;
-  
-  if (!pk_algo_table[i].is_ecc
-      && strcmp (pk_algo_table[i].elem_string, "-ne"))
-    return NULL; /* Not ECC or RSA - we can't handle it yet.  */
 
-  *r_is_ecc = pk_algo_table[i].is_ecc;
+  *r_pkalgo = pk_algo_table[i].pkalgo;
   *oidlen = pk_algo_table[i].oidlen;
   return pk_algo_table[i].oid;
 }
@@ -817,19 +877,23 @@ _ksba_keyinfo_from_sexp (ksba_const_sexp_t sexp,
   int oidlen;
   unsigned char *curve_oid = NULL;
   size_t curve_oidlen;
-  int is_ecc;
+  pkalgo_t pkalgo;
   int i;
   struct {
     const char *name;
     int namelen;
     const unsigned char *value;
     int valuelen;
-  } parm[3];
+  } parm[10]; 
   int parmidx;
+  int idxtbl[10];
+  int idxtbllen;
+  const char *parmdesc, *algoparmdesc;
   ksba_writer_t writer = NULL;
+  void *algoparmseq_value = NULL;
+  size_t algoparmseq_len;
   void *bitstr_value = NULL;
   size_t bitstr_len;
-  int ecc_curve_idx, ecc_q_idx;  
 
   if (!sexp)
     return gpg_error (GPG_ERR_INV_VALUE);
@@ -857,7 +921,7 @@ _ksba_keyinfo_from_sexp (ksba_const_sexp_t sexp,
   if (!n || *s != ':')
     return gpg_error (GPG_ERR_INV_SEXP); /* we don't allow empty lengths */
   s++;
-  oid = oid_from_buffer (s, n, &oidlen, &is_ecc);
+  oid = oid_from_buffer (s, n, &oidlen, &pkalgo);
   if (!oid)
     return gpg_error (GPG_ERR_UNSUPPORTED_ALGORITHM);
   s += n;
@@ -898,39 +962,56 @@ _ksba_keyinfo_from_sexp (ksba_const_sexp_t sexp,
   if ( *s != ')' )
     return gpg_error (GPG_ERR_INV_SEXP); 
 
-  ecc_q_idx = ecc_curve_idx = -1;
-  if (is_ecc)
+  /* Describe the parameters in the order we want them and construct
+     IDXTBL to access them.  For DSA wie also set algoparmdesc so
+     that we can later build the parameters for the
+     algorithmIdentifier.  */
+  algoparmdesc = NULL;
+  switch (pkalgo)
+    {
+    case PKALGO_RSA: parmdesc = "ne"; break;
+    case PKALGO_DSA: parmdesc = "y" ; algoparmdesc = "pqg"; break;
+    case PKALGO_ECC: parmdesc = "Cq"; break;
+    default: return gpg_error (GPG_ERR_UNKNOWN_ALGORITHM);
+    }      
+
+  idxtbllen = 0;
+  for (s = parmdesc; *s; s++)
     {
       for (i=0; i < parmidx; i++)
         {
-          if (ecc_curve_idx == -1 
-              && parm[i].namelen == 5 && !memcmp (parm[i].name, "curve", 5))
-            ecc_curve_idx = i;
-          else if (ecc_q_idx == -1 
-              && parm[i].namelen == 1 && parm[i].name[0] == 'q')
-            ecc_q_idx = i;
+          assert (idxtbllen < DIM (idxtbl));
+          switch (*s)
+            {
+            case 'C': /* Magic value for "curve".  */
+              if (parm[i].namelen == 5 && !memcmp (parm[i].name, "curve", 5))
+                {
+                  idxtbl[idxtbllen++] = i;
+                  i = parmidx; /* Break inner loop.  */
+                }
+              break;
+            default:
+              if (parm[i].namelen == 1 && parm[i].name[0] == *s)
+                {
+                  idxtbl[idxtbllen++] = i;
+                  i = parmidx; /* Break inner loop.  */
+                }
+              break;
+            }
         }
-      if (ecc_curve_idx == -1 || ecc_q_idx == -1)
-        return gpg_error (GPG_ERR_UNKNOWN_SEXP);
+    }
+  if (idxtbllen != strlen (parmdesc))
+    return gpg_error (GPG_ERR_UNKNOWN_SEXP);
 
-      curve_oid = get_ecc_curve_oid (parm[ecc_curve_idx].value,
-                                     parm[ecc_curve_idx].valuelen,
+  if (pkalgo == PKALGO_ECC)
+    {
+      curve_oid = get_ecc_curve_oid (parm[idxtbl[0]].value,
+                                     parm[idxtbl[0]].valuelen,
                                      &curve_oidlen);
       if (!curve_oid)
         return gpg_error (GPG_ERR_UNKNOWN_SEXP);
     }
-  else /* This is RSA.  */
-    {
-      /* Check that the names match the requirements for RSA */
-      s = "ne"; 
-      if (parmidx != strlen (s))
-        return gpg_error (GPG_ERR_UNKNOWN_SEXP);
-      for (i=0; i < parmidx; i++)
-        {
-          if (parm[i].namelen != 1 || parm[i].name[0] != s[i])
-            return gpg_error (GPG_ERR_UNKNOWN_SEXP);
-        }
-    }      
+
       
   /* Create write object. */
   err = ksba_writer_new (&writer);
@@ -940,42 +1021,43 @@ _ksba_keyinfo_from_sexp (ksba_const_sexp_t sexp,
   if (err)
     goto leave;
   
-  if (is_ecc)
+  /* We create the keyinfo in 2 steps:
+     
+     1. We build the inner one and encapsulate it in a bit string.
+  
+     2. We create the outer sequence include the algorithm identifier
+        and the bit string from step 1. 
+   */
+  if (pkalgo == PKALGO_ECC)
     {
       /* Write the bit string header and the number of unused bits. */
       err = _ksba_ber_write_tl (writer, TYPE_BIT_STRING, CLASS_UNIVERSAL, 
-                                0, parm[ecc_q_idx].valuelen + 1);
+                                0, parm[idxtbl[1]].valuelen + 1);
       if (!err)
         err = ksba_writer_write (writer, "", 1);
       /* And the actual raw value.  */
       if (!err)
-        err = ksba_writer_write (writer, parm[ecc_q_idx].value,
-                                 parm[ecc_q_idx].valuelen);
+        err = ksba_writer_write (writer, parm[idxtbl[1]].value,
+                                 parm[idxtbl[1]].valuelen);
       if (err)
         goto leave;
 
     }
-  else /* RSA */
+  else /* RSA and DSA */
     {
-      /* For RSA we create the keyinfo in 2 steps:
-
-        1. We build the inner one and encapsulate it in bit string.
-        
-        2. We create the outer sequence include the algorithm identifier
-           and the bit string from step 1.  
-      */
-
-      /* Calculate the size of the sequence value and the size of the bit
-         string value */
-      for (n=0, i=0; i < parmidx; i++ )
+      /* Calculate the size of the sequence value and the size of the
+         bit string value.  NOt ethat in case there is only one
+         integer to write, no sequence is used.  */
+      for (n=0, i=0; i < idxtbllen; i++ )
         {
           n += _ksba_ber_count_tl (TYPE_INTEGER, CLASS_UNIVERSAL, 0,
-                                   parm[i].valuelen);
-          n += parm[i].valuelen;
+                                   parm[idxtbl[i]].valuelen);
+          n += parm[idxtbl[i]].valuelen;
         }
   
       n1 = 1; /* # of unused bits.  */
-      n1 += _ksba_ber_count_tl (TYPE_SEQUENCE, CLASS_UNIVERSAL, 1, n);
+      if (idxtbllen > 1)
+        n1 += _ksba_ber_count_tl (TYPE_SEQUENCE, CLASS_UNIVERSAL, 1, n);
       n1 += n;
   
       /* Write the bit string header and the number of unused bits. */
@@ -987,17 +1069,19 @@ _ksba_keyinfo_from_sexp (ksba_const_sexp_t sexp,
         goto leave;
       
       /* Write the sequence tag and the integers. */
-      err = _ksba_ber_write_tl (writer, TYPE_SEQUENCE, CLASS_UNIVERSAL, 1, n);
+      if (idxtbllen > 1)
+        err = _ksba_ber_write_tl (writer, TYPE_SEQUENCE, CLASS_UNIVERSAL, 1,n);
       if (err)
         goto leave;
-      for (i=0; i < parmidx; i++)
+      for (i=0; i < idxtbllen; i++)
         {
           /* fixme: we should make sure that the integer conforms to the
              ASN.1 encoding rules. */
           err  = _ksba_ber_write_tl (writer, TYPE_INTEGER, CLASS_UNIVERSAL, 0, 
-                                     parm[i].valuelen);
+                                     parm[idxtbl[i]].valuelen);
           if (!err)
-            err = ksba_writer_write (writer, parm[i].value, parm[i].valuelen);
+            err = ksba_writer_write (writer, parm[idxtbl[i]].value, 
+                                     parm[idxtbl[i]].valuelen);
           if (err)
             goto leave;
         }
@@ -1010,6 +1094,66 @@ _ksba_keyinfo_from_sexp (ksba_const_sexp_t sexp,
       err = gpg_error (GPG_ERR_ENOMEM);
       goto leave;
     }
+  
+  /* If the algorithmIdentifier requires a sequence with parameters,
+     build them now.  We can reuse the IDXTBL for that.  */
+  if (algoparmdesc)
+    {
+      idxtbllen = 0;
+      for (s = algoparmdesc; *s; s++)
+        {
+          for (i=0; i < parmidx; i++)
+            {
+              assert (idxtbllen < DIM (idxtbl));
+              if (parm[i].namelen == 1 && parm[i].name[0] == *s)
+                {
+                  idxtbl[idxtbllen++] = i;
+                  break;
+                }
+            }
+        }
+      if (idxtbllen != strlen (algoparmdesc))
+        return gpg_error (GPG_ERR_UNKNOWN_SEXP);
+
+      err = ksba_writer_set_mem (writer, 1024);
+      if (err)
+        goto leave;
+
+      /* Calculate the size of the sequence.  */
+      for (n=0, i=0; i < idxtbllen; i++ )
+        {
+          n += _ksba_ber_count_tl (TYPE_INTEGER, CLASS_UNIVERSAL, 0,
+                                   parm[idxtbl[i]].valuelen);
+          n += parm[idxtbl[i]].valuelen;
+        }
+      //      n += _ksba_ber_count_tl (TYPE_SEQUENCE, CLASS_UNIVERSAL, 1, n);
+  
+      /* Write the sequence tag followed by the integers. */
+      err = _ksba_ber_write_tl (writer, TYPE_SEQUENCE, CLASS_UNIVERSAL, 1, n);
+      if (err)
+        goto leave;
+      for (i=0; i < idxtbllen; i++)
+        {
+          err = _ksba_ber_write_tl (writer, TYPE_INTEGER, CLASS_UNIVERSAL, 0, 
+                                    parm[idxtbl[i]].valuelen);
+          if (!err)
+            err = ksba_writer_write (writer, parm[idxtbl[i]].value, 
+                                     parm[idxtbl[i]].valuelen);
+          if (err)
+            goto leave;
+        }
+
+      /* Get the encoded sequence.  */
+      algoparmseq_value = ksba_writer_snatch_mem (writer, &algoparmseq_len);
+      if (!algoparmseq_value)
+        {
+          err = gpg_error (GPG_ERR_ENOMEM);
+          goto leave;
+        }
+    }
+  else
+    algoparmseq_len = 0;
+
   /* Reinitialize the buffer to create the outer sequence. */
   err = ksba_writer_set_mem (writer, 1024);
   if (err)
@@ -1018,13 +1162,17 @@ _ksba_keyinfo_from_sexp (ksba_const_sexp_t sexp,
   /* Calulate lengths. */
   n  = _ksba_ber_count_tl (TYPE_OBJECT_ID, CLASS_UNIVERSAL, 0, oidlen);
   n += oidlen;
-  if (is_ecc)
+  if (algoparmseq_len)
+    {
+      n += algoparmseq_len;
+    }
+  else if (pkalgo == PKALGO_ECC)
     {
       n += _ksba_ber_count_tl (TYPE_OBJECT_ID, CLASS_UNIVERSAL,
                                0, curve_oidlen);
       n += curve_oidlen;
     }
-  else
+  else if (pkalgo == PKALGO_RSA)
     {
       n += _ksba_ber_count_tl (TYPE_NULL, CLASS_UNIVERSAL, 0, 0);
     }
@@ -1051,14 +1199,18 @@ _ksba_keyinfo_from_sexp (ksba_const_sexp_t sexp,
     goto leave;
 
   /* The parameter. */
-  if (is_ecc)
+  if (algoparmseq_len)
+    {
+      err = ksba_writer_write (writer, algoparmseq_value, algoparmseq_len);
+    }
+  else if (pkalgo == PKALGO_ECC)
     {
       err = _ksba_ber_write_tl (writer, TYPE_OBJECT_ID, CLASS_UNIVERSAL,
                                 0, curve_oidlen);
       if (!err)
         err = ksba_writer_write (writer, curve_oid, curve_oidlen);
     }
-  else /* RSA */
+  else if (pkalgo == PKALGO_RSA)
     {
       err = _ksba_ber_write_tl (writer, TYPE_NULL, CLASS_UNIVERSAL, 0, 0);
     }
@@ -1091,7 +1243,7 @@ cryptval_to_sexp (int mode, const unsigned char *der, size_t derlen,
                   ksba_sexp_t *r_string)
 {
   gpg_error_t err;
-  struct algo_table_s *algo_table;
+  const struct algo_table_s *algo_table;
   int c;
   size_t nread, off, len;
   int algoidx;
@@ -1165,9 +1317,9 @@ cryptval_to_sexp (int mode, const unsigned char *der, size_t derlen,
             return gpg_error (GPG_ERR_INV_KEYINFO);
           c = *der++; derlen--;
           if ( c != *ctrl )
-            return gpg_error (GPG_ERR_UNEXPECTED_TAG); /* not the required tag */
+            return gpg_error (GPG_ERR_UNEXPECTED_TAG);
           is_int = c == 0x02;
-          TLV_LENGTH ();
+          TLV_LENGTH (der);
         }
       if (is_int && *elem != '-')
         { /* take this integer */
@@ -1182,7 +1334,15 @@ cryptval_to_sexp (int mode, const unsigned char *der, size_t derlen,
           put_stringbuf (&sb, ")");
         }
     }
-  put_stringbuf (&sb, "))");
+  put_stringbuf (&sb, ")");
+  if (!mode && algo_table[algoidx].digest_string)
+    {
+      /* Insert the hash algorithm if included in the OID.  */
+      put_stringbuf (&sb, "(4:hash");
+      put_stringbuf_sexp (&sb, algo_table[algoidx].digest_string);
+      put_stringbuf (&sb, ")");
+    }
+  put_stringbuf (&sb, ")");
   
   *r_string = get_stringbuf (&sb);
   if (!*r_string)
@@ -1208,8 +1368,8 @@ cryptval_to_sexp (int mode, const unsigned char *der, size_t derlen,
     (<algo>
       (<param_name1> <mpi>)
       ...
-      (<param_namen> <mpi>)
-    ))
+      (<param_namen> <mpi>))
+    (hash algo))
 
  The S-Exp will be returned in a string which the caller must free.
  We don't pass an ASN.1 node here but a plain memory block.  */
